"""Module provider for Misaka.IO"""

from __future__ import absolute_import

import base64
import json
import logging
from argparse import ArgumentParser
from typing import List, Tuple

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


def _recordset_has_record(record_set, value):
    for record in record_set["records"]:
        if record["value"] == value:
            return True
    return False


class Provider(BaseProvider):
    """Provider class for Misaka.IO"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return [
            # gTLDs
            "m1ns.com",
            "m1ns.net",
            "m1ns.org",
            # new gTLDs
            "m1ns.one",
            "m1ns.moe",
            "m1ns.xyz",
            "m1ns.fyi",
            "m1ns.app",
            # ccTLDs
            "m1ns.be",
            "m1ns.io",
            "m1ns.uk",
            "m1ns.us",
            "m1ns.im",
            # for PTR zones only
            "reversedns.org",
            # legacy domains
            "ns53.net",
        ]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify token for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.misaka.io/dns"

    def authenticate(self):
        payload = self._get(f"/zones/{self.domain}/settings")
        if not payload["id"]:
            raise AuthenticationError("No domain found")

        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    # Create or update a record.
    def create_record(self, rtype, name, content):
        name = self._relative_name(name)

        identifier = self._identifier_encode(rtype, name)
        endpoint = f"/zones/{self.domain_id}/recordsets/{name}/{rtype}"
        ttl = self._get_lexicon_option("ttl")
        record = {"value": content}

        existing_recordset = self._get_recordset(name, rtype)
        if existing_recordset:
            # append if returned recordsets doesn't include the record
            if not _recordset_has_record(existing_recordset, content):
                existing_recordset["records"].append(record)
                existing_recordset["ttl"] = ttl
                self._put(endpoint, existing_recordset)
                LOGGER.debug("recordset exists, appending: %s", identifier)
            # update ttl if returned ttl doesn't match
            elif existing_recordset["ttl"] != ttl:
                existing_recordset["ttl"] = ttl
                self._put(endpoint, existing_recordset)
                LOGGER.debug("recordset exists, updating ttl: %s", identifier)
        else:
            recordset = {
                "ttl": ttl,
                "filters": [],
                "records": [record],
            }
            self._post(endpoint, recordset)
            LOGGER.debug("create_record: %s", identifier)

        return True

    def list_records(self, rtype=None, name=None, content=None):
        params = {}
        if name:
            name = self._relative_name(name)
            params["name"] = name

        payload = self._get(f"/zones/{self.domain_id}/recordsets", query_params=params)
        records = []
        for recordset in payload["results"]:
            if not recordset["records"]:
                continue

            if name and recordset["name"] != name:
                continue

            if rtype and recordset["type"] != rtype:
                continue

            processed_recordset = {
                "name": self._full_name(recordset["name"]),
                "type": recordset["type"],
                "ttl": recordset["ttl"],
                "id": self._identifier_encode(recordset["type"], recordset["name"]),
            }

            for record in recordset["records"]:
                processed_record = {"content": record["value"]}
                processed_record.update(processed_recordset)
                records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        # use relative_name only
        if name:
            name = self._relative_name(name)

        new_identifier = self._identifier_encode(rtype, name)

        if new_identifier == identifier or (rtype is None and name is None):
            # the identifier hasn't changed, or type and name are both unspecified,
            # only update the content.
            data = {"records": {"value": content}}
            target_rtype, target_name = self._identifier_decode(identifier)
            self._put(
                f"/zones/{self.domain_id}/recordsets/{target_name}/{target_rtype}", data
            )
        else:
            if not identifier:
                identifier = new_identifier
            # identifiers are different
            # get the old record, create a new one with updated data, delete the old record.
            target_rtype, target_name = self._identifier_decode(identifier)
            old_record = self._get(
                f"/zones/{self.domain_id}/recordsets/{target_name}/{target_rtype}"
            )
            self.create_record(
                rtype or old_record["type"],
                name or old_record["domain"],
                content or old_record["records"][0]["value"],
            )
            self.delete_record(identifier)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        should_call_delete_api = True
        if not identifier:
            if name:
                name = self._relative_name(name)
            identifier = self._identifier_encode(rtype, name)
            should_call_delete_api = not self._delete_record_with_identifier(
                identifier, rtype, name, content
            )

        if should_call_delete_api:
            rtype, name = self._identifier_decode(identifier)
            self._delete(f"/zones/{self.domain_id}/recordsets/{name}/{rtype}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Delete an existing record if identifier isn't None.
    # it will return True if we don't need to call delete API
    def _delete_record_with_identifier(self, identifier, rtype, name, content):
        recordset = self._get_recordset(name, rtype)
        if not recordset:
            return True

        # remove corresponding record only if content isn't None
        if not content:
            return False
        recordset["records"] = [
            record for record in recordset["records"] if record["value"] != content
        ]

        if not recordset["records"]:
            return False

        rtype, name = self._identifier_decode(identifier)
        self._put(f"/zones/{self.domain_id}/recordsets/{name}/{rtype}", recordset)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        # set defaults
        data = {} if data is None else data
        query_params = {} if query_params is None else query_params

        token = self._get_provider_option("auth_token")

        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Token {token}",
        }
        default_auth = None

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _get_recordset(self, name, rtype):
        try:
            payload = self._get(f"/zones/{self.domain_id}/recordsets/{name}/{rtype}")
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            raise

        return {
            "ttl": payload["ttl"],
            "records": payload["records"],
            "filters": payload["filters"],
        }

    def _identifier_decode(self, identifier: str) -> Tuple[str, str]:
        padding = 4 - (len(identifier) % 4)
        chain = identifier + ("=" * padding)
        decoded = base64.urlsafe_b64decode(chain)
        if not decoded:
            raise ValueError(f"Invalid identifier: {identifier}")
        extracted = decoded.decode("utf-8").split("/")
        if len(extracted) < 2:
            raise ValueError(f"Invalid identifier: {identifier}")
        return extracted[0], extracted[1]

    def _identifier_encode(self, rtype: str, name: str) -> str:
        encoded = base64.urlsafe_b64encode(
            f"{rtype}/{self._relative_name(name)}".encode("utf-8")
        )
        return encoded.decode("utf-8").rstrip("=")



"""
Notes on this Provider

1. Make sure the NIOS Version support WAPI 2.6.1
2. Have a valid Certificate from a public CA installed at the Infoblox

Commandline examples:
1. all parameters given:
lexicon infoblox --ib-host myib.mydomain.tld --auth-user {username} --auth-psw {passwordd}
    --ib-view default create test.local A --content 10.10.10.11 --name lexicon1
2. Parameters mixed with ENV
LEXICON_INFOBLOX_AUTH_USER={user} LEXICON_INFOBLOX_AUTH_PSW={password} lexicon infoblox
    --ib-host myib.mydomain.tld --ib-view default create test.local A
    --content 10.10.10.11 --name lexicon1

"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# SOA and NS are not record types itself, but rather auto created depending on the Zone settings.
# Either Primary Grid Member directly assigned to the zone, or through a ns_group
# skipping for now
"""
Dictionary to map the record Type to different Infoblox specific values
IB_TYPE2CONTENT = {
    'type':['WAPI atribute field content','WAPI base URL per type','WAPI returnfields'],
}
"""
IB_TYPE2CONTENT = {
    "A": ["ipv4addr", "record:a", ",ttl,use_ttl"],
    "AAAA": ["ipv6addr", "record:aaaa", ",ttl,use_ttl"],
    "CNAME": ["canonical", "record:cname", ",ttl,use_ttl"],
    "MX": ["mail_exchanger", "record:mx", ",ttl,use_ttl"],
    "TXT": ["text", "record:txt", ",ttl,use_ttl"],
    "SRV": ["target", "record:srv", ",ttl,use_ttl"],
}

_NAMESERVER_DOMAINS = ["test.local."]


class Provider(BaseProvider):
    """Provider class for Infoblox"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return _NAMESERVER_DOMAINS

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-user", help="specify the user to access the Infoblox WAPI"
        )
        parser.add_argument(
            "--auth-psw", help="specify the password to access the Infoblox WAPI"
        )
        parser.add_argument(
            "--ib-view",
            default="default",
            help="specify DNS View to manage at the Infoblox",
        )
        parser.add_argument("--ib-host", help="specify Infoblox Host exposing the WAPI")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        # In Case .local. Domains are used, ignore the tldextract in client.py
        # this will be used in test_infoblox.py as well
        if self.config.resolve("lexicon:domain") in _NAMESERVER_DOMAINS[0]:
            self.domain = _NAMESERVER_DOMAINS[0].rstrip(".")
        self.domain_id = None
        self.version_id = None
        self.view = self._get_provider_option("ib_view")
        self.session = requests.session()
        self.session.auth = (
            self._get_provider_option("auth_user"),
            self._get_provider_option("auth_psw"),
        )
        self.version = "2.6.1"  # WAPI version supported by NIOS 8.3 and above
        self.session.headers.update({"Content-Type": "application/json"})
        self.api_endpoint = (
            f"https://{self._get_provider_option('ib_host')}/wapi/v{self.version}/"
        )

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def authenticate(self):
        response = self.session.get(
            f"{self.api_endpoint}zone_auth?fqdn={self.domain}&view={self.view}"
        )
        domains = response.json()
        try:
            self.domain_id = domains[0]["_ref"]
        except IndexError:
            LOGGER.error("Domain %s not found in view", self.domain)
            raise AuthenticationError(f"Domain {self.domain} not found in view")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        if name:
            name = self._fqdn_name(name)
        else:
            raise Exception("Name not specified, no FQDN could be build")
        # Find existing records for all types
        existing = []
        for rrtype in IB_TYPE2CONTENT:
            existing = existing + self.list_records(rtype=rrtype, name=name)
        # we don't want to delete all existing A,AAAA,TXT,SRV,MX,NS records
        # which can not co-exist with CNAMEs
        if any(d["type"] == rtype for d in existing):
            # no conflict in types
            if any(d["content"] == content and d["type"] == rtype for d in existing):
                # already exists
                return True
            if rtype == "CNAME":
                # we found a CNAME entry; we update it with the new target
                return self._update_record_internal(
                    existing[0]["id"], rtype, name, content
                )
            return self._create_record_internal(rtype, name, content)
        if any(d["type"] == "CNAME" and rtype != "CNAME" for d in existing):
            # we found a CNAME entry; we can delete it and create the requested type
            if self.delete_record(identifier=existing[0]["id"]):
                return self._create_record_internal(rtype, name, content)
            LOGGER.error("Deleting record failed for:%s", existing[0]["id"])
            return False
        if any(d["type"] != "CNAME" and rtype == "CNAME" for d in existing):
            LOGGER.error("CNAME requested, but other records already exists")
            raise Exception("CNAME requested, but other records already exists")
        return self._create_record_internal(rtype, name, content)

    def _create_record_internal(self, rtype, name, content):
        if name:
            name = self._fqdn_name(name)
        else:
            raise Exception("Name not specified, no FQDN could be build")
        if rtype.upper() in IB_TYPE2CONTENT:
            uri = self.api_endpoint + IB_TYPE2CONTENT[rtype.upper()][1]
            payload = self._generate_payload(rtype, name, content)
            try:
                response = self.session.post(uri, data=json.dumps(payload))
            except BaseException:
                LOGGER.error("Connection Error during create")
                raise Exception("Connection Error")
            if response.status_code == 201:
                return True
            return False
        raise Exception("RR Type not supported by Infoblox")

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        # Infoblox stores entries based on their type, if type is not specified look up all types
        if not rtype:
            records = []
            for one_rtype in IB_TYPE2CONTENT:
                records = records + self._list_records_internal(
                    rtype=one_rtype, name=name, content=content
                )
            return records
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(self, rtype=None, name=None, content=None):
        # infoblox expect the fullname when looking it up without trailing dot
        if name:
            name = self._fqdn_name(name)
        else:
            name = self.domain
        records = []
        if rtype.upper() in IB_TYPE2CONTENT:
            uri = "{0}{1}?name={2}&view={3}&_return_fields={4}{5}".format(
                self.api_endpoint,
                IB_TYPE2CONTENT[rtype.upper()][1],
                name,
                self.view,
                IB_TYPE2CONTENT[rtype.upper()][0],
                IB_TYPE2CONTENT[rtype.upper()][2],
            )
            try:
                response = self.session.get(uri)
            except BaseException:
                raise Exception("Connection Error")

            results = response.json()
            for result in results:
                if content:
                    # return exact match
                    if (
                        str(content).lower()
                        == str(result[IB_TYPE2CONTENT[rtype][0]]).lower()
                    ):
                        if result["use_ttl"]:
                            record = {
                                "type": rtype,
                                "name": name,
                                "ttl": result["ttl"],
                                "content": result[IB_TYPE2CONTENT[rtype][0]],
                                "id": result["_ref"],
                            }
                        else:
                            record = {
                                "type": rtype,
                                "name": name,
                                "ttl": 3600,  # replace by Infoblox default TTL
                                "content": result[IB_TYPE2CONTENT[rtype][0]],
                                "id": result["_ref"],
                            }
                        records.append(record)
                else:
                    # return any record
                    if result["use_ttl"]:
                        record = {
                            "type": rtype,
                            "name": name,
                            "ttl": result["ttl"],
                            "content": result[IB_TYPE2CONTENT[rtype][0]],
                            "id": result["_ref"],
                        }
                    else:
                        record = {
                            "type": rtype,
                            "name": name,
                            "ttl": 3600,  # replace by Infoblox default TTL
                            "content": result[IB_TYPE2CONTENT[rtype][0]],
                            "id": result["_ref"],
                        }
                    records.append(record)
            return records
        return records

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            return self._update_record_internal(
                identifier=identifier, rtype=rtype, name=name, content=content
            )
        success = []
        for record in self.list_records(rtype, name, content):
            success.append(
                self._update_record_internal(
                    identifier=record["id"], rtype=rtype, name=name, content=content
                )
            )
        if False in success:
            return False
        return True

    def _update_record_internal(self, identifier, rtype=None, name=None, content=None):
        uri = self.api_endpoint + identifier
        payload = self._generate_payload(rtype, name, content)
        # remove view and name from the payload, as they can not be updated
        del payload["view"]
        del payload["name"]
        try:
            response = self.session.put(uri, data=json.dumps(payload))
        except BaseException:
            LOGGER.error("Connection Error during create")
            raise Exception("Connection Error")
        if response.status_code == 200:
            return True
        return False

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Infoblox Object Identifier example:
        # record:cname/ZG5zLmJpbmRfY25hbWUkLjIzLmRlLm1naS5sZXhpY29uNA:lexicon.domain.tld/view
        if identifier:
            return self._delete_record_internal(identifier)
        success = []
        for record in self.list_records(rtype, name, content):
            success.append(self._delete_record_internal(record["id"]))
        if False in success:
            return False
        return True

    def _delete_record_internal(self, identifier=None):
        if identifier:
            uri = f"{self.api_endpoint}{identifier}"
            try:
                response = self.session.delete(uri)
            except BaseException:
                LOGGER.error("Connection Error during create")
                raise Exception("Connection Error")
            if response.status_code == 200:
                return True
            return False
        return False

    # Helpers
    def _generate_payload(self, rtype, name, content):
        # build full object as required by Infoblox WAPI
        payload = {}
        payload["view"] = self.view
        payload["name"] = name
        if rtype.upper() == "TXT":
            payload[IB_TYPE2CONTENT[rtype.upper()][0]] = '"' + content + '"'
        else:
            payload[IB_TYPE2CONTENT[rtype.upper()][0]] = content
        if self._get_lexicon_option("ttl"):
            payload["ttl"] = self._get_lexicon_option("ttl")
            payload["use_ttl"] = True
        # MX and SRV have additional fields
        if rtype.upper() == "MX":
            payload["preference"] = self._get_lexicon_option("priority")
        elif rtype.upper() == "SRV":
            payload["priority"] = self._get_lexicon_option("priority")
            payload["weight"] = self._get_provider_option("ib_weight")
            payload["port"] = self._get_provider_option("ib_port")
        return payload

    def _fqdn_name(self, record_name):
        record_name = record_name.rstrip(
            "."
        )  # strip trailing period from fqdn if present
        # check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = f"{record_name}.{self.domain}"
        return f"{record_name}"  # return the fqdn name without trailing dot

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in Infoflox provider
        pass



"""Module provider for Namesilo"""

import logging
from argparse import ArgumentParser
from typing import List
from xml.etree import ElementTree

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Namesilo"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["namesilo.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify key for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint") or "https://www.namesilo.com/api"
        )

    def authenticate(self):
        try:
            self._get("/getDomainInfo", {"domain": self.domain})
        except Exception as e:
            raise AuthenticationError(str(e))
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'

    def create_record(self, rtype, name, content):
        record = {
            "domain": self.domain_id,
            "rrhost": self._relative_name(name),
            "rrtype": rtype,
            "rrvalue": content,
        }
        if self._get_lexicon_option("ttl"):
            record["rrttl"] = self._get_lexicon_option("ttl")
        try:
            self._get("/dnsAddRecord", record)
        except ValueError as err:
            # noop if attempting to create record that already exists.
            LOGGER.debug("Ignoring error: %s", err)

        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        query = {"domain": self.domain_id}

        payload = self._get("/dnsListRecords", query)
        records = []
        for record in payload.find("reply").findall("resource_record"):
            processed_record = {
                "type": record.find("type").text,
                "name": record.find("host").text,
                "ttl": record.find("ttl").text,
                "content": record.find("value").text,
                "id": record.find("record_id").text,
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {"domain": self.domain_id, "rrid": identifier}
        # if rtype:
        #     data['type'] = rtype
        if name:
            data["rrhost"] = self._relative_name(name)
        if content:
            data["rrvalue"] = content
        if self._get_lexicon_option("ttl"):
            data["rrttl"] = self._get_lexicon_option("ttl")

        self._get("/dnsUpdateRecord", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        data = {"domain": self.domain_id}

        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            data["rrid"] = record_id
            self._get("/dnsDeleteRecord", data)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        query_params["version"] = 1
        query_params["type"] = "xml"
        query_params["key"] = self._get_provider_option("auth_token")
        response = requests.request(
            action, self.api_endpoint + url, params=query_params
        )
        # data=json.dumps(data))
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        tree = ElementTree.ElementTree(ElementTree.fromstring(response.content))
        root = tree.getroot()
        if root.find("reply").find("code").text == "280":
            raise ValueError(
                "An error occurred: {0}, {1}".format(
                    root.find("reply").find("detail").text,
                    root.find("reply").find("code").text,
                )
            )
        if root.find("reply").find("code").text != "300":
            raise Exception(
                "An error occurred: {0}, {1}".format(
                    root.find("reply").find("detail").text,
                    root.find("reply").find("code").text,
                )
            )

        return root



"""Parsers definition for the Lexicon command-line interface"""

import argparse
import os
from typing import Type

from lexicon._private.discovery import (
    find_providers,
    lexicon_version,
    load_provider_module,
)
from lexicon.interfaces import Provider


def generate_base_provider_parser() -> argparse.ArgumentParser:
    """Function that generates the base provider to be used by all dns providers."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "action",
        help="specify the action to take",
        default="list",
        choices=["create", "list", "update", "delete"],
    )
    parser.add_argument(
        "domain", help="specify the domain, supports subdomains as well"
    )
    parser.add_argument(
        "type",
        help="specify the entry type",
        default="TXT",
        choices=["A", "AAAA", "CNAME", "MX", "NS", "SOA", "TXT", "SRV", "LOC"],
    )

    parser.add_argument("--name", help="specify the record name")
    parser.add_argument("--content", help="specify the record content")
    parser.add_argument("--ttl", type=int, help="specify the record time-to-live")
    parser.add_argument("--priority", help="specify the record priority")
    parser.add_argument(
        "--identifier", help="specify the record for update or delete actions"
    )
    parser.add_argument(
        "--log-level",
        "--log_level",
        help="specify the log level",
        default="ERROR",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )
    parser.add_argument(
        "--output",
        help=(
            "specify the type of output: by default a formatted table (TABLE), "
            "a formatted table without header (TABLE-NO-HEADER), "
            "a JSON string (JSON) or no output (QUIET)"
        ),
        default="TABLE",
        choices=["TABLE", "TABLE-NO-HEADER", "JSON", "QUIET"],
    )
    return parser


def generate_cli_main_parser() -> argparse.ArgumentParser:
    """Using all providers available, generate a parser that will be used by Lexicon CLI"""
    parser = argparse.ArgumentParser(
        description="Create, Update, Delete, List DNS entries"
    )

    parser.add_argument(
        "--version",
        help="show the current version of lexicon",
        action="version",
        version=f"%(prog)s {lexicon_version()}",
    )
    parser.add_argument(
        "--delegated",
        help="specify the delegated domain (may not needed if --resolve-zone-name is set)",
    )
    parser.add_argument(
        "--config-dir",
        default=os.getcwd(),
        help="specify the directory where to search lexicon.yml and "
        "lexicon_[provider].yml configuration files "
        "(default: current directory).",
    )
    parser.add_argument(
        "--resolve-zone-name",
        action="store_true",
        help="trigger an active resolution of the zone name for the given domain using DNS queries",
    )
    subparsers = parser.add_subparsers(
        dest="provider_name", help="specify the DNS provider to use"
    )
    subparsers.required = True

    for provider, available in find_providers().items():
        provider_module = load_provider_module(provider)
        provider_class: Type[Provider] = getattr(provider_module, "Provider")

        subparser = subparsers.add_parser(
            provider,
            help=f"{provider} provider",
            parents=[generate_base_provider_parser()],
        )

        provider_class.configure_parser(subparser)

        if not available:
            subparser.epilog = (
                "WARNING: some required dependencies for this provider are not "
                f"installed. Please run `pip install lexicon[{provider}]` first before using it."
            )

    return parser



"""Module provider for Vultr"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Vultr"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["vultr.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify token for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.vultr.com/v2"

    def authenticate(self):
        payload = self._get("/domains")

        for domain in payload["domains"]:
            if domain["domain"] == self.domain:
                self.domain_id = self.domain
                return

        while payload["meta"]["links"]["next"] != "":
            query_params = {"cursor": payload["meta"]["links"]["next"]}
            payload = self._get("/domains", query_params=query_params)

            for domain in payload["domains"]:
                if domain["domain"] == self.domain:
                    self.domain_id = self.domain
                    return

        raise AuthenticationError("Domain not found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        records = self.list_records(rtype, name, content)
        if len(records) != 0:
            LOGGER.debug("create_record (already exists): %s", records[0]["id"])
            return True

        record = {
            "type": rtype,
            "name": self._relative_name(name),
            "data": self._add_quotes(rtype, content),
            "priority": 0,
        }
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        result = self._post(f"/domains/{self.domain_id}/records", record)
        LOGGER.debug("create_record: %s", result["record"]["id"])
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        url = f"/domains/{self.domain_id}/records"

        payload = self._get(url)
        unprocessed_records = payload["records"]

        while payload["meta"]["links"]["next"] != "":
            query_params = {"cursor": payload["meta"]["links"]["next"]}
            payload = self._get(url, query_params=query_params)
            unprocessed_records.extend(payload["records"])

        records = []
        for record in unprocessed_records:
            records.append(self._process_record(record))

        if rtype:
            records = [rec for rec in records if rec["type"] == rtype]
        if name:
            records = [rec for rec in records if rec["name"] == self._full_name(name)]
        if content:
            records = [rec for rec in records if rec["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        record = None
        if not identifier:
            records = self.list_records(rtype, name)

            if not records:
                raise Exception(
                    f"No record(s) found for arguments: identifer={identifier}, rtype={rtype}, name={name}"
                )
            if len(records) > 1:
                LOGGER.warning(
                    "Multiple records have been found for given parameters. "
                    "Only first one will be updated (id: %s)",
                    records[0]["id"],
                )

            record = records[0]
            identifier = record["id"]

        url = f"/domains/{self.domain_id}/records/{identifier}"
        if not record:
            record = self._get(url)["record"]
            record = self._process_record(record)

        new_record = {}

        if name:
            name = self._relative_name(name)
            if name != record["name"]:
                new_record["name"] = name

        if content:
            content = self._add_quotes(record["type"], content)
            if content != record["content"]:
                new_record["data"] = content

        if new_record == {}:
            LOGGER.debug("update_record (nothing to do): %s", True)
            return True

        self._patch(url, new_record)
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)
        for record_id in delete_record_id:
            try:
                self._delete(f"/domains/{self.domain_id}/records/{record_id}")
            except requests.HTTPError as e:
                if e.response.status_code != 404:
                    raise

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {
            "Accept": "application/json",
            "Authorization": "Bearer " + self._get_provider_option("auth_token"),
        }

        if data is not None:
            headers["Content-Type"] = "application/json"
            data = json.dumps(data)

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        if response.status_code == 204:
            return None

        return response.json()

    @staticmethod
    def _add_quotes(rtype, content):
        if rtype == "TXT":
            return f'"{content}"'
        return content

    def _process_record(self, record):
        processed_record = {
            "type": record["type"],
            "name": self._full_name(record["name"]),
            "ttl": record["ttl"],
            "content": record["data"],
            "id": record["id"],
        }
        return self._clean_TXT_record(processed_record)



"""Module provider for DNS.services"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for DNS.services"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dns.services"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.auth_token = None
        self.domain_id = None
        self.api_endpoint = "https://dns.services/api"

    def authenticate(self):
        if self.auth_token is None:
            username = self._get_provider_option("auth_username")
            password = self._get_provider_option("auth_password")
            data = {
                "username": username,
                "password": password,
            }

            result = requests.post(self.api_endpoint + "/login", data=data)
            result.raise_for_status()
            self.auth_token = result.json()["token"]

        data = self._get("/dns")
        for zone in data["zones"]:
            if zone["name"] == self.domain:
                self.domain_id = zone["domain_id"]
                self.service_id = zone["service_id"]
                return
        raise AuthenticationError("No domain found")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        print("create_record")
        # check if record already exists
        if not self.list_records(rtype, name, content):
            data = {
                "type": rtype,
                "name": self._relative_name(name),
                "content": content,
            }
            ttl = self._get_lexicon_option("ttl")
            if ttl:
                data["ttl"] = ttl
            priority = self._get_lexicon_option("priority")
            if priority:
                data["priority"] = str(priority)
            LOGGER.debug("create_record: %s", data)
            self._post(f"/service/{self.service_id}/dns/{self.domain_id}/records", data)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        print("list_records")
        payload = self._get(f"/service/{self.service_id}/dns/{self.domain_id}")
        records = []
        for _, record in payload["records"].items():
            processed_record = {
                "type": record["type"],
                "name": record["name"],
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        print("update_record")
        data = {
            "type": rtype,
            "name": self._relative_name(name),
            "content": content,
        }
        ttl = self._get_lexicon_option("ttl")
        if ttl:
            data["ttl"] = ttl
        priority = self._get_lexicon_option("priority")
        if priority:
            data["priority"] = str(priority)
        LOGGER.debug("update_record: %s", data)
        self._put(
            f"/service/{self.service_id}/dns/{self.domain_id}/records/{identifier}",
            data,
        )
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(
                f"/service/{self.service_id}/dns/{self.domain_id}/records/{record_id}"
            )

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        print("request")
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.auth_token}",
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        return response.json()



"""
===============================
Module provider for ValueDomain
===============================


Value-Domain API
================

Reference: https://www.value-domain.com/vdapi/

- Preparation

1. Get your auth token: https://www.value-domain.com/vdapi/
2. export LEXICON_VALUEDOMAIN_AUTH_TOKEN="<Your Auth Token>"
3. export DOMAIN_NAME=$(curl -s https://api.value-domain.com/v1/domains -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -c "import sys,json;print(json.load(sys.stdin)['results'][0]['domainname'])")

- Get list of your domain information

curl -s https://api.value-domain.com/v1/domains -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool

- Get domain name

curl -s https://api.value-domain.com/v1/domains -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -c "import sys,json;print(json.load(sys.stdin)['results'][0]['domainname'])"

- Get name servers

curl -s https://api.value-domain.com/v1/domains/nameserver -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool
curl -s https://api.value-domain.com/v1/domains/${DOMAIN_NAME}/nameserver -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool

- Get DNS records

curl -s https://api.value-domain.com/v1/domains/dns -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool
curl -s https://api.value-domain.com/v1/domains/${DOMAIN_NAME}/dns -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool

- Get API Logs

curl -s https://api.value-domain.com/v1/logs -H "Authorization: Bearer ${LEXICON_VALUEDOMAIN_AUTH_TOKEN}" | python -m json.tool


Value Doamin DNS records syntax
===============================

Reference: https://www.value-domain.com/moddnsfree.php

- A record
  - a www 123.123.123.3                : A record for www.<YOUR-DOMAIN>
  - a * 123.123.123.5                  : A record for any sub domain of <YOUR-DOMAIN>
  - a @ 123.123.123.5                  : A record for <YOUR-DOMAIN> (no sub domain)

- AAAA record
  - aaaa ipv6 FF01::101                : AAAA record for ipv6.<YOUR-DOMAIN>

- MX record
  - mx mx1.your.domain. 10             : MX record for <YOUR-DOMAIN> (Server:mx1.your.domain, Priority 10)
  - mx @ 10                            : MX record for <YOUR-DOMAIN> (Server:<YOUR-DOMAIN>, Priority 10)

- NS record
  - ns abc ns1.example.com.            : NS record for abc.<YOUR-DOMAIN>

- TXT record
  - txt abc v=spf1 mx ~all             : TXT record for abc.<YOUR-DOMAIN> (v=spf1 mx ~all)
  - txt @ v=spf1 mx ~all               : TXT record for <YOUR-DOMAIN> (v=spf1 mx ~all)

- SRV record
  - srv _smtp._tcp 1 2 25 server1.your.domain   : SRV record for _smtp._tcp.<YOUR-DOMAIN> (Priority:1, Weight:2, Port:25)
"""

from __future__ import annotations

import hashlib
import json
import logging
import time
from argparse import ArgumentParser
from http import cookiejar
from http.client import HTTPResponse
from typing import Any, Callable, Dict, List, NamedTuple, Optional, TypeVar, Union
from urllib import request
from urllib.error import HTTPError
from urllib.request import OpenerDirector

from lexicon.config import ConfigResolver
from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

DEFAULT_TTL = 3600

########################################################################
# Util
########################################################################


def convert_json_to_bytes(x):
    return bytes(json.dumps(x), "utf-8")


def is_domain(target: str) -> bool:
    return (
        len(target) > 1
        and len(target.strip(".")) > 0
        and target[0] != "."
        and target[-1] == "."
    )


def is_sub_domain(target: str, domainname: str) -> Union[str, bool]:
    idx = target.rfind(domainname)
    return (
        target[0:idx].strip(".")
        if idx >= 0 and is_domain(target) and idx + len(domainname) + 1 == len(target)
        else False
    )


########################################################################
# Rest API
########################################################################


class RestApiResponse(NamedTuple):
    header: HTTPResponse
    data: bytes


T = TypeVar("T")
RestAPICallable = Callable[[str, str, T], RestApiResponse]


def reastapi_add_content_type(
    _headers: Optional[dict[str, str]],
    content_type: str | None,
    content: Optional[bytes],
) -> dict[str, str]:
    """Add 'Content-Type' header if exists"""
    headers = _headers.copy() if _headers is not None else {}
    if content_type is not None and content is not None and len(content) != 0:
        headers["Content-Type"] = content_type
    return headers


def restapi_create_request(
    url: str,
    method: str,
    headers: dict[str, str] | None,
    content_type: str | None,
    content: bytes | None,
) -> request.Request:
    """Create Request instance including content if exists"""
    return request.Request(
        url,
        data=content if content is not None and len(content) > 0 else None,
        method=method,
        headers=reastapi_add_content_type(headers, content_type, content),
    )


def restapi_call(
    opener: OpenerDirector,
    url: str,
    method: str,
    headers: dict[str, str] | None,
    content_type: str | None = None,
    content: bytes | None = None,
) -> RestApiResponse:
    """Execute HTTP Request with OpenerDirector"""
    with opener.open(
        restapi_create_request(url, method, headers, content_type, content)
    ) as response:
        return RestApiResponse(response, response.read())


def restapi_build_opener() -> OpenerDirector:
    """Create OpenerDirector instance with cookie processor"""
    return request.build_opener(
        request.BaseHandler(), request.HTTPCookieProcessor(cookiejar.CookieJar())
    )


def restapi_exception_not_200(response: HTTPResponse):
    # HTTPError will be raised if status >= 400
    if response.status != 200:
        LOGGER.error(f"HTTP Status: {response.status}")
        raise Exception(f"HTTP Status: {response.status}")


########################################################################
# Value Domain API
########################################################################


VDAPI_ENDPOINT = "https://api.value-domain.com/v1"


class RecordData(NamedTuple):
    rtype: str
    name: str
    content: str

    def __eq__(self, other) -> bool:
        return (
            self.rtype.lower() == other.rtype.lower()
            and self.name.lower() == other.name.lower()
            and self.content == other.content
        )

    def __str__(self) -> str:
        return f"{self.rtype.lower()} {self.name.lower()} {self.content}"

    def match(
        self,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> bool:
        return (
            (rtype is None or rtype.lower() == self.rtype.lower())
            and (name is None or name.lower() == self.name.lower())
            and (content is None or content.lower() == self.content.lower())
        )

    def id(self):
        return hashlib.md5(str(self).encode("utf-8")).hexdigest()


class DomainData(NamedTuple):
    records: List[RecordData]
    ttl: int


def vdapi_build_caller(
    opener: OpenerDirector,
    content_type: str,
    headers: dict[str, str] | None = None,
    content_decoder=lambda x: x,
) -> RestAPICallable:
    def _(
        url: str, method: str, content: Optional[T] = None, interval=1
    ) -> RestApiResponse:
        try:
            return restapi_call(
                opener,
                url,
                method,
                headers,
                content_type,
                content_decoder(content) if content is not None else None,
            )
        except HTTPError as http_error:
            if http_error.code == 429:  # Too much Request
                time.sleep(interval)
                return _(url, method, content, interval * 2)
            else:
                raise http_error

    return _


def vdapi_create_caller(auth_token: str) -> RestAPICallable:
    return vdapi_build_caller(
        restapi_build_opener(),
        "application/json",
        {
            # Value-Domain API rejects the request with default Python User-Agent.
            "User-Agent": "curl/7.74.0",
            "Cache-Control": "no-cache, no-store",
            "Authorization": f"Bearer {auth_token}",
        },
        convert_json_to_bytes,
    )


def vdapi_get_domain_list(caller: RestAPICallable) -> list[str]:
    resp: RestApiResponse = caller(f"{VDAPI_ENDPOINT}/domains", "GET", None)
    restapi_exception_not_200(resp.header)
    return list(
        filter(
            lambda x: x is not None,
            [
                domain.get("domainname")
                for domain in json.loads(resp.data.decode("utf-8").strip()).get(
                    "results"
                )
            ],
        )
    )


def vdapi_get_domain_data(
    caller: RestAPICallable, domainname: str
) -> DomainData | None:
    resp: RestApiResponse = caller(
        f"{VDAPI_ENDPOINT}/domains/{domainname}/dns", "GET", None
    )
    restapi_exception_not_200(resp.header)
    domain_info: dict = json.loads(resp.data.decode("utf-8").strip())
    domain_records: str | None = domain_info.get("results", {}).get("records")
    return (
        DomainData(
            [
                RecordData(elem[0], elem[1], elem[2])
                for elem in [line.split(" ", 3) for line in domain_records.split("\n")]
                if len(elem) == 3
            ],
            int(domain_info.get("results", {}).get("ttl", DEFAULT_TTL)),
        )
        if domain_records is not None
        else None
    )


def vdapi_set_domain_data(
    caller: RestAPICallable, domainname: str, data: DomainData
) -> None:
    resp = caller(
        f"{VDAPI_ENDPOINT}/domains/{domainname}/dns",
        "PUT",
        {
            "ns_type": "valuedomain1",
            "records": "\n".join([str(record) for record in data.records]),
            "ttl": data.ttl,
        },
    )
    restapi_exception_not_200(resp.header)


########################################################################
# Lexicon Provider for Value-Domain
########################################################################


class Provider(BaseProvider):
    """Provider class for Value Domain"""

    @staticmethod
    def get_nameservers() -> list[str]:
        return ["value-domain.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            Value Domain requires a token to access its API.
            You can generate one for your account on the following URL:
            https://www.value-domain.com/vdapi/"""
        parser.add_argument("--auth-token", help="specify youyr API token")

    def __init__(self, config: ConfigResolver):
        super(Provider, self).__init__(config)

        auth_token = self._get_provider_option("auth_token")
        assert auth_token is not None, "No authenticaion token defined"

        self.caller: RestAPICallable = vdapi_create_caller(auth_token)

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def authenticate(self):
        domains = vdapi_get_domain_list(self.caller)

        assert len(domains) > 0, "Failed to get domain names"
        if self.domain not in domains:
            raise AuthenticationError(f"{self.domain} not managed")

        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype: str, name: str, content: str):
        self._assert_initialized()
        ttl_option = self._get_lexicon_option("ttl")
        domain_data = vdapi_get_domain_data(self.caller, self.domain)
        rec = self._create_record_data(rtype, self._relative_name(name), content)

        if domain_data is not None and rec not in domain_data.records:
            vdapi_set_domain_data(
                self.caller,
                self.domain,
                DomainData(
                    domain_data.records + [rec],
                    (
                        int(ttl_option)
                        if ttl_option is not None and int(ttl_option) > 0
                        else DEFAULT_TTL if ttl_option is not None else domain_data.ttl
                    ),
                ),
            )
        elif domain_data is not None:
            pass
        else:
            vdapi_set_domain_data(
                self.caller,
                self.domain,
                DomainData(
                    [rec],
                    (
                        int(ttl_option)
                        if ttl_option is not None and int(ttl_option) > 0
                        else DEFAULT_TTL
                    ),
                ),
            )

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(
        self,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ):
        self._assert_initialized()
        domain_data = vdapi_get_domain_data(self.caller, self.domain)
        return (
            [
                {
                    "id": record_data.id(),
                    "ttl": domain_data.ttl,
                    "type": record_data.rtype.upper(),
                    "name": (
                        self._fqdn_name(record_data.name)
                        if record_data.rtype.lower() != "txt"
                        else self._full_name(record_data.name)
                    ),
                    "content": record_data.content,
                }
                for record_data in domain_data.records
                if record_data.match(
                    rtype,
                    self._relative_name(name) if name is not None else None,
                    content,
                )
            ]
            if domain_data is not None
            else []
        )

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        self._assert_initialized()
        ttl_option = self._get_lexicon_option("ttl")
        domain_data = vdapi_get_domain_data(self.caller, self.domain)
        target = [record for record in domain_data.records if record.id() == identifier]

        if len(target) > 0:
            vdapi_set_domain_data(
                self.caller,
                self.domain,
                DomainData(
                    [
                        record
                        for record in domain_data.records
                        if record.id() != identifier
                    ]
                    + [
                        self._create_record_data(
                            rtype or target[0].rtype,
                            name or self._relative_name(target[0].name),
                            content or target[0].content,
                        )
                    ],
                    (
                        ttl_option
                        if ttl_option is not None and ttl_option > 0
                        else DEFAULT_TTL if ttl_option is not None else domain_data.ttl
                    ),
                ),
            )

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        self._assert_initialized()
        domain_data = vdapi_get_domain_data(self.caller, self.domain)
        if domain_data is not None and identifier is not None:
            vdapi_set_domain_data(
                self.caller,
                self.domain,
                DomainData(
                    [
                        record
                        for record in domain_data.records
                        if record.id() != identifier
                    ],
                    domain_data.ttl,
                ),
            )
        elif domain_data is not None:
            vdapi_set_domain_data(
                self.caller,
                self.domain,
                DomainData(
                    [
                        record
                        for record in domain_data.records
                        if not record.match(rtype, self._relative_name(name), content)
                    ],
                    domain_data.ttl,
                ),
            )

        return True

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: Optional[Dict] = None,
        query_params: Optional[Dict] = None,
    ) -> Any:
        pass

    #        self._assert_initialized()
    #        return self.caller(url, action, data)

    def _assert_initialized(self):
        if self.caller is None or self.domain_id is None:
            self.authenticate()

        assert self.caller is not None, "HTTP caller not defined"
        assert self.domain_id is not None or len(
            self.domain_id == 0
        ), "Domain name not retriebed"

    def _create_record_data(self, rtype: str, name: str, content: str) -> RecordData:
        return RecordData(rtype.lower(), self._relative_name(name).lower(), content)


if __name__ == "__main__":
    import os
    import unittest

    DUMMY_TYPE = "a"
    DUMMY_NAME = "test"
    DUMMY_CONTENT = "1.2.3.4"
    DUMMY_CONTENT2 = "2.3.4.5"

    class TestValueDomain(unittest.TestCase):
        def setUp(self):
            self.auth_token = os.environ.get("LEXICON_VALUEDOMAIN_AUTH_TOKEN")
            if self.auth_token is None:
                raise Exception("LEXICON_VALUEDOMAIN_AUTH_TOKEN not defined")
            self.caller = vdapi_create_caller(self.auth_token)

        def tearDown(self):
            pass

        def _create_provider(self, domainname: str):
            config = ConfigResolver().with_dict(
                {
                    "provider_name": "valuedomain",
                    "domain": domainname,
                    "valuedomain": {"auth_token": self.auth_token},
                }
            )
            return Provider(config)

        def test_vdapi_get_domain_list(self):
            domain_list = vdapi_get_domain_list(self.caller)
            self.assertGreater(len(domain_list), 0)

        def test_vdapi_get_record_list(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                record = vdapi_get_domain_data(self.caller, domainname)

                self.assertIsNotNone(record)
                self.assertGreater(len(record.records), 0)
                self.assertIsNotNone(record.ttl)

        def test_vdapi_set_record_list(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                domain_data = vdapi_get_domain_data(self.caller, domainname)
                dummy = RecordData(DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                vdapi_set_domain_data(
                    self.caller,
                    domainname,
                    DomainData(domain_data.records + [dummy], domain_data.ttl),
                )
                updated_domain_data = vdapi_get_domain_data(self.caller, domainname)
                self.assertIn(dummy, updated_domain_data.records)

                # cleanup
                vdapi_set_domain_data(self.caller, domainname, domain_data)
                updated_domain_data = vdapi_get_domain_data(self.caller, domainname)
                self.assertNotIn(dummy, updated_domain_data.records)

        def test_list_records(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                provider = self._create_provider(domainname)
                provider.authenticate()

                records = provider.list_records()
                self.assertGreater(len(records), 0)

                records = provider.list_records(rtype="A")
                self.assertGreater(len(records), 0)

                records = provider.list_records(rtype="B")
                self.assertEqual(len(records), 0)

        def test_create_records(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                provider = self._create_provider(domainname)
                provider.authenticate()

                provider.create_record(DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                records = provider.list_records(
                    rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                )
                self.assertGreater(len(records), 0)

                provider.create_record(
                    DUMMY_TYPE, f"{DUMMY_NAME}.{domainname}.", DUMMY_CONTENT
                )
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE,
                            name=f"DUMMY_NAME.{domainname}.",
                            content=DUMMY_CONTENT,
                        )
                    ),
                    0,
                )

                # cleanup
                provider.delete_record(None, DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                        )
                    ),
                    0,
                )

        def test_delete_records_by_id(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                provider = self._create_provider(domainname)
                provider.authenticate()

                provider.create_record(DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                recl = provider.list_records(
                    rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                )
                self.assertGreater(len(recl), 0)
                provider.delete_record(identifier=recl[0].get("id"))
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                        )
                    ),
                    0,
                )

        def test_delete_records_by_data(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                provider = self._create_provider(domainname)
                provider.authenticate()

                provider.create_record(DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                recl = provider.list_records(
                    rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                )
                self.assertGreater(len(recl), 0)
                provider.delete_record(None, DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                        )
                    ),
                    0,
                )

        def test_update_record(self):
            domain_list = vdapi_get_domain_list(self.caller)
            for domainname in domain_list:
                provider = self._create_provider(domainname)
                provider.authenticate()

                provider.create_record(DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT)
                recl = provider.list_records(
                    rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                )
                self.assertGreater(len(recl), 0)
                provider.update_record(
                    identifier=recl[0].get("id"), content=DUMMY_CONTENT2
                )
                self.assertGreater(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT2
                        )
                    ),
                    0,
                )

                # cleanup
                provider.delete_record(None, DUMMY_TYPE, DUMMY_NAME, DUMMY_CONTENT2)
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT
                        )
                    ),
                    0,
                )
                self.assertEqual(
                    len(
                        provider.list_records(
                            rtype=DUMMY_TYPE, name=DUMMY_NAME, content=DUMMY_CONTENT2
                        )
                    ),
                    0,
                )

    unittest.main()



"""
The Constellix API has some limitations.  We try to paper over them here,
but here's what you need to be aware of:

 1) SOA records are not first-class record types in the Constellix API, so are not supported.
 2) We expect all records to use the "Standard" record type, so failover, pools or round robin with
    failover are not supported.
 3) Because Constellix represents record sets as a single record with multiple values attached,
    not as a set of separate records, create and delete operations end up becoming read/update
    operations when working with record sets.

    Since these aren't atomic operations, it creates a small window where you could have data loss
    if multiple processes were trying to work with the same record set.

    This is unlikely to be a problem in most scenarios, but the possilbity is there.  I've reached
    out to the Constellix folks to see if they have plans to clean up the API to resolve this.
"""

import base64
import hashlib
import hmac
import json
import logging
import time
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider clss for Constellix"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["constellix.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify the API key username for authentication"
        )
        parser.add_argument("--auth-token", help="specify secret key for authenticate=")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.domain_details = None
        self.api_endpoint = "https://api.dns.constellix.com/v1"

    def authenticate(self):
        try:
            payload = self._get("/domains/")
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                payload = {}
            else:
                raise error

        for domain in payload:
            if domain["name"] == self.domain:
                self.domain_id = domain["id"]
                self.domain_details = domain
                break
        else:
            raise AuthenticationError("No domain found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'

    def create_record(self, rtype, name, content):
        record = {
            "name": self._relative_name(name),
            "ttl": self._get_lexicon_option("ttl"),
            "roundRobin": [{"disableFlag": False, "value": content}],
        }
        payload = {}

        try:
            payload = self._post(f"/domains/{self.domain_id}/records/{rtype}/", record)
        except requests.exceptions.HTTPError as error:
            # If there is already a record with that name, we need to do an update.
            if error.response.status_code == 400:
                existing_records = self.list_records(rtype=rtype, name=name)
                new_content = [r["content"] for r in existing_records]

                # Only do the update if we are creating a record that doesn't already exist,
                # otherwise Constellix will throw an error.
                if content not in new_content:
                    new_content.append(content)
                    self.update_record(
                        existing_records[0]["id"],
                        rtype=rtype,
                        name=name,
                        content=new_content,
                    )
            else:
                raise
        LOGGER.debug("create_record: %s", "name" in payload)
        return True

    # Currently returns the first value for hosts where there may be multiple
    # values.  Need to check to see how this is handled for other providers.
    def list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        self._check_type(rtype)

        # Oddly, Constellix supports API-level filtering for everything except LOC
        # records, so we need to retrieve all records for LOC and filter based on rtype
        # on our end.
        if not rtype or rtype == "LOC":
            payload = self._get(f"/domains/{self.domain_id}/records/")
        else:
            payload = self._get(f"/domains/{self.domain_id}/records/{rtype}/")

        records = []

        for record in payload:
            for a_record in record["roundRobin"]:
                processed_record = {
                    "type": record["type"],
                    "name": f"{record['name']}.{self.domain}",
                    "ttl": record["ttl"],
                    "content": a_record["value"],
                    "id": record["id"],
                }

                processed_record = self._clean_TXT_record(processed_record)
                records.append(processed_record)

        records = self._filter_records(
            records, rtype=rtype, name=name, content=content, identifier=identifier
        )

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        self._check_type(rtype)

        if content and not isinstance(content, (list)):
            content = [content]

        if identifier and (not rtype or not name):
            record = self._list_records_internal(identifier=identifier)
            rtype = record[0]["type"]
            name = record[0]["name"]
        elif not identifier:
            record = self.list_records(rtype, name)
            identifier = record[0]["id"]

        if not identifier:
            raise Exception("No identifier provided")

        data = {
            "id": identifier,
            "ttl": self._get_lexicon_option("ttl"),
            "name": self._relative_name(name),
        }

        data["roundRobin"] = []

        for a_content in content:
            data["roundRobin"].append({"disableFlag": False, "value": a_content})

        self._put(f"/domains/{self.domain_id}/records/{rtype}/{identifier}/", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        self._check_type(rtype)

        records = self._list_records_internal(
            identifier=identifier, rtype=rtype, name=name
        )

        # If we are filtering delete records by content and we are going to have
        # at least one record left over after deleting, then this becomes an
        # update operation.
        if content:
            current_content = set(r["content"] for r in records)
            if content in current_content and len(current_content) > 1:
                current_content.remove(content)
                self.update_record(
                    records[0]["id"],
                    rtype=rtype,
                    name=name,
                    content=list(current_content),
                )
                return True

        delete_record_id = set(record["id"] for record in records)

        # We need a rtype to do a delete, so pull one from the first record if it's not supplied.
        if not rtype:
            rtype = records[0]["type"]

        for record_id in delete_record_id:
            self._delete(f"/domains/{self.domain_id}/records/{rtype}/{record_id}/")

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _check_type(self, rtype=None):
        # Constellix doesn't treat SOA as a separate record type, so we bail on SOA modificiations.
        # It looks like it would be possible to fake SOA CRUD, so an area for possible future
        # improvement

        if rtype == "SOA":
            raise Exception(
                f"{rtype} record type is not supported in the Constellix Provider"
            )

        return True

    def _filter_records(
        self, records, rtype=None, name=None, content=None, identifier=None
    ):
        _records = []
        for record in records:
            if (
                (not identifier or record["id"] == identifier)
                and (not rtype or record["type"] == rtype)
                and (not name or record["name"] == self._full_name(name))
                and (not content or record["content"] == content)
            ):
                _records.append(record)
        return _records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-cnsdns-apiKey": self._get_provider_option("auth_username"),
        }
        default_auth = None

        # Date string in epoch format
        request_date = str(int(time.time() * 1000))

        hashed = hmac.new(
            self._get_provider_option("auth_token").encode("utf-8"),
            request_date.encode("utf-8"),
            digestmod=hashlib.sha1,
        )

        default_headers["x-cnsdns-requestDate"] = request_date
        default_headers["x-cnsdns-hmac"] = base64.b64encode(hashed.digest())

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()

        # PUT and DELETE actions dont return valid json.
        if action in ["DELETE" or action == "PUT"]:
            return response.text

        return response.json()



"""Module provider for Gehirn"""

import base64
import json
import logging
import re
from argparse import ArgumentParser
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

BUILD_FORMATS = {
    "A": "{address}",
    "AAAA": "{address}",
    "CNAME": "{cname}",
    "TXT": "{data}",
    "NS": "{nsdname}",
    "MX": "{prio} {exchange}",
    "SRV": "{prio} {weight} {port} {target}",
}

FORMAT_RE = {
    "A": re.compile(r"(?P<address>.+)"),
    "AAAA": re.compile(r"(?P<address>.+)"),
    "CNAME": re.compile(r"(?P<cname>.+)"),
    "TXT": re.compile(r"(?P<data>.+)"),
    "NS": re.compile(r"(?P<nsdname>.+)"),
    "MX": re.compile(r"(?P<prio>\d+)\s+(?P<exchange>.+)"),
    "SRV": re.compile(
        r"(?P<prio>\d+)\s+(?P<weight>\d+)\s+(?P<port>\d+)\s+(?P<target>.+)"
    ),
}


class Provider(BaseProvider):
    """Provider class for Gehirn"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["gehirn.jp"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token", help="specify access token for authentication"
        )
        parser.add_argument(
            "--auth-secret", help="specify access secret for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.version_id = None
        self.api_endpoint = "https://api.gis.gehirn.jp/dns/v1"

    def authenticate(self):
        payload = self._get("/zones")

        domains = [item for item in payload if item["name"] == self.domain]
        if not domains:
            raise AuthenticationError("No domain found")

        self.domain_id = domains[0]["id"]
        self.version_id = domains[0]["current_version_id"]

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        name = self._full_name(name)
        a_record = self._parse_content(rtype, content)

        record = None
        records = self._get_records(rtype=rtype, name=name)
        if not records:
            record = {
                "type": rtype,
                "name": name,
                "enable_alias": False,
                "ttl": self._get_lexicon_option("ttl"),
                "records": [],
            }
        else:
            record = records[0]

        if a_record in record["records"]:
            LOGGER.debug("create_record: %s", True)
            return True

        record["records"].append(a_record)
        self._update_internal_record(record)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        records = []
        if name:
            name = self._full_name(name)
        for record in self._get_records(rtype=rtype, name=name):
            for a_record in record["records"]:
                content = self._build_content(record["type"], a_record)
                processed_record = {
                    "type": record["type"],
                    "name": record["name"].rstrip("."),
                    "ttl": record["ttl"],
                    "content": content,
                    "id": f"{record['id']}.{base64.b64encode(content.encode('utf-8')).decode('ascii')}",
                }
                self._parse_content(record["type"], processed_record["content"])
                records.append(processed_record)

        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if name:
            name = self._full_name(name)

        if not identifier:
            if not (rtype and name and content):
                raise Exception("type, name and content must be specified.")

            records = self._get_records(rtype=rtype, name=name)

            if not records:
                self.create_record(rtype=rtype, name=name, content=content)
                LOGGER.debug("update_record: %s", True)
                return True

            record = {
                "id": records[0]["id"],
                "type": rtype,
                "name": name,
                "enable_alias": False,
                "ttl": self._get_lexicon_option("ttl"),
                "records": [self._parse_content(rtype, content)],
            }

        else:
            # with identifier
            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception("Record identifier could not be found.")

            record = records[0]

            if "." in identifier:
                # modify single record
                self.delete_record(identifier=identifier)
                self.create_record(
                    rtype=rtype or record["type"],
                    name=name or record["name"],
                    content=content,
                )
                LOGGER.debug("update_record: %s", True)
                return True
            # update entire record
            if rtype:
                record["type"] = rtype
            if name:
                record["name"] = name
            record["ttl"] = self._get_lexicon_option("ttl")
            if content:
                record["records"] = [self._parse_content(record["type"], content)]

        self._update_internal_record(record)
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            if "." not in identifier:
                # delete entire record
                path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{identifier}"
                self._delete(path)
                LOGGER.debug("delete_record: %s", True)
                return True

            record_identifier = identifier.split(".")[1]

            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception("Record identifier could not be found.")

            record = records[0]
            for index, a_record in enumerate(record["records"]):
                target_content = self._build_content(record["type"], a_record)
                target_identifier = base64.b64encode(
                    target_content.encode("utf-8")
                ).decode("ascii")

                if target_identifier == record_identifier:
                    del record["records"][index]
                if not record["records"]:
                    # delete entire record
                    path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{record['id']}"
                    self._delete(path)
                else:
                    self._update_internal_record(record)

                LOGGER.debug("delete_record: %s", True)
                return True

            raise Exception("Record identifier could not be found.")

        record = None
        if name is not None:
            name = self._full_name(name)
        if content is not None:
            content = self._bind_format_target(rtype, content)
            record = self._parse_content(rtype, content)

        records = self._get_records(rtype=rtype, name=name)

        for a_record in records:
            if record and record in a_record["records"]:
                a_record["records"].remove(record)
                if a_record["records"]:
                    self._update_internal_record(a_record)
                    continue

            path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{a_record['id']}"
            self._delete(path)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _full_name(self, record_name):
        record_name = super(Provider, self)._full_name(record_name)
        if not record_name.endswith("."):
            record_name += "."
        return record_name

    def _bind_format_target(self, rtype, target):
        if rtype == "CNAME" and not target.endswith("."):
            target += "."
        return target

    def _filter_records(self, records, identifier=None, rtype=None, name=None):
        filtered_records = []

        if identifier:
            identifier = identifier.split(".")[0]

        for record in records:
            if rtype and record["type"] != rtype:
                continue
            if name and record["name"] != name:
                continue
            if identifier and record["id"] != identifier:
                continue
            filtered_records.append(record)
        return filtered_records

    def _get_records(self, identifier=None, rtype=None, name=None):
        path = f"/zones/{self.domain_id}/versions/{self.version_id}/records"
        return self._filter_records(
            self._get(path), identifier=identifier, rtype=rtype, name=name
        )

    def _update_internal_record(self, record):
        if record.get("id"):
            # PUT
            path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{record['id']}"
            return self._put(path, record)

        # POST
        path = f"/zones/{self.domain_id}/versions/{self.version_id}/records"
        return self._post(path, record)

    def _build_content(self, rtype, record):
        return BUILD_FORMATS[rtype].format(**record)

    def _parse_content(self, rtype, content):
        return FORMAT_RE[rtype].match(content).groupdict()

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        default_auth = HTTPBasicAuth(
            self._get_provider_option("auth_token"),
            self._get_provider_option("auth_secret"),
        )

        query_string = ""
        if query_params:
            query_string = json.dumps(query_params)

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_string,
            data=json.dumps(data) if data else None,
            headers=default_headers,
            auth=default_auth,
        )
        try:
            # if the request fails for any reason, throw an error.
            response.raise_for_status()
        except BaseException:
            LOGGER.error(response.text)
            raise
        return response.json()



"""Module provider for DDNS"""

import hashlib
import logging
from argparse import ArgumentParser
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

# dnspython is an optional dependency of lexicon; do not throw an ImportError if
# the dependency is unmet.
try:
    import dns.message
    import dns.query
    import dns.tsigkeyring
    import dns.update
    import dns.versioned
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for DDNS"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token",
            help="specify the key used in format <alg>:<key_id>:<secret>",
        )
        parser.add_argument("--ddns-server", help="specify IP of the DDNS server")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        alg, keyid, secret = self._get_provider_option("auth_token").split(":")
        self.keyring = dns.tsigkeyring.from_text({keyid: (alg, secret)})
        self.endpoint = self._get_provider_option("ddns_server")
        self.zone = self._get_provider_option("domain")
        self.cached_zone_content = {}

    def _run_query(self, message):
        return dns.query.tcp(message, self.endpoint, timeout=10)

    def authenticate(self):
        if not self.endpoint:
            raise AuthenticationError("No DDNS server provided, use --ddns-server")
        pass

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        if self._get_lexicon_option("ttl"):
            ttl = self._get_lexicon_option("ttl")
        else:
            ttl = 300
        name = dns.name.from_text(name).relativize(dns.name.from_text(self.zone))

        update = dns.update.Update(self.zone, keyring=self.keyring)
        update.add(name, ttl, rtype, content)
        self._run_query(update)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        if rtype and name:  # doing a simple query is enough if we have type and name
            query = dns.message.make_query(name, rtype)
        else:  # if not, perform a zone transfert to get all records
            query = dns.xfr.make_query(
                dns.versioned.Zone(self.zone), keyring=self.keyring
            )[0]
            self.cached_zone_content = {}

        answers = self._run_query(query).answer

        records = []
        for answer in answers:  # filter unwanted results
            a_name = answer.name.to_text().rstrip(".")
            if name and name != a_name:
                continue
            a_rtype = dns.rdatatype.to_text(answer.rdtype)
            if rtype and rtype != a_rtype:
                continue
            for rdata in answer:
                rdata = rdata.to_text()
                if content and content != rdata:
                    continue
                identifier = _identifier(a_rtype, a_name, rdata)
                record = {
                    "type": a_rtype,
                    "name": a_name,
                    "ttl": answer.ttl,
                    "content": rdata,
                    "id": identifier,
                }
                records.append(record)
                self.cached_zone_content[identifier] = (a_rtype, a_name, rdata)
        LOGGER.debug("list_records: %s", records)
        LOGGER.debug("Number of records retrieved: %d", len(records))
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if self._get_lexicon_option("ttl"):
            ttl = self._get_lexicon_option("ttl")
        else:
            ttl = 300
        if not identifier:
            rrset = self.list_records(rtype, name)
            if len(rrset) == 1:
                identifier = rrset[0]["id"]
            elif len(rrset) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )
        d_rtype, d_name, d_content = self._resolve_identifier(identifier)

        if not rtype:
            rtype = d_rtype
        if not name:
            name = d_name

        d_name = dns.name.from_text(d_name).relativize(dns.name.from_text(self.zone))
        name = dns.name.from_text(name).relativize(dns.name.from_text(self.zone))

        update = dns.update.Update(self.zone, keyring=self.keyring)
        update.delete(d_name, d_rtype, d_content)
        update.add(name, ttl, rtype, content)
        self._run_query(update)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            rtype, name, content = self._resolve_identifier(identifier)

        name = dns.name.from_text(name).relativize(dns.name.from_text(self.zone))
        update = dns.update.Update(self.zone, keyring=self.keyring)
        if content:
            update.delete(name, rtype, content)
        else:
            update.delete(name, rtype)
        self._run_query(update)

        return True

    def _resolve_identifier(self, ident):
        if not self.cached_zone_content.get(ident):
            self.list_records()  # if cache miss, reload cache
        return self.cached_zone_content[ident]

    def _request(self, action="GET", url="/", data=None, query_params=None):
        pass


def _identifier(record):
    m = hashlib.sha256()
    m.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    m.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    m.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

    return m.hexdigest()[0:7]



"""Module provider for Dreamhost"""

import base64
import json
import logging
import time
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

_DATA_NON_EXIST_ERROR_LIST = [
    "no_record",
    "no_type",
    "no_value",
    "no_such_record",
    "no_such_type",
    "no_such_value",
    "no_such_zone",
]

_DATA_ALREADY_EXIST_ERROR_LIST = [
    "record_already_exists_not_editable",
    "record_already_exists_remove_first",
    "CNAME_already_on_record",
]


class NonExistError(Exception):
    """NonExistError"""


class AlreadyExistError(Exception):
    """AlreadyExistError"""


class Provider(BaseProvider):
    """Provider class for Dreamhost"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dreamhost.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api key for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.dreamhost.com/"

    # Dreamhost provides no identifier for a record.
    # Furthermore, Dreamhost requires type, record, value to delete a record.
    # The record defined in lexicon is {type, name, content, id}
    # We use base64(json({'type', 'name', 'content'}))
    # as the identifier of Dreamhost record.
    @staticmethod
    def _identifier(dreamhost_record):
        id_struct = {
            "type": dreamhost_record["type"],
            "name": dreamhost_record["record"],
            "content": dreamhost_record["value"],
        }
        return base64.urlsafe_b64encode(json.dumps(id_struct).encode("utf-8")).decode(
            "utf-8"
        )

    # The information in identifier follows the record in lexicon.
    # Provider._record_to_dreamhost_record transfers to dreamhost-based record.
    @staticmethod
    def _id_to_dreamhost_record(identifier):
        record = json.loads(
            base64.urlsafe_b64decode(identifier.encode("utf-8")).decode("utf-8")
        )
        dreamhost_record = Provider._record_to_dreamhost_record(record)
        return dreamhost_record

    # The information in identifier follows the record in lexicon.
    # 'id' is added in the record.
    @staticmethod
    def _id_to_record(identifier):
        record = json.loads(
            base64.urlsafe_b64decode(identifier.encode("utf-8")).decode("utf-8")
        )
        record["id"] = identifier

        return record

    # Transferring lexicon-based record to Dreamhost-based record.
    @staticmethod
    def _record_to_dreamhost_record(record):
        dreamhost_record = {
            "type": record["type"],
            "record": record["name"],
            "value": record["content"],
        }
        return dreamhost_record

    def authenticate(self):
        self.domain_id = None
        payload = self._get("dns-list_records")
        data = payload.get("data", None)
        if data is None:
            raise AuthenticationError("Domain not found")

        for record in data:
            if record.get("record", "") == self.domain and record.get("type", "") in [
                "A",
                "AAAA",
                "CNAME",
                "MX",
                "NS",
                "SOA",
                "TXT",
                "SRV",
            ]:
                self.domain_id = self.domain
                break
        if self.domain_id is None:
            raise AuthenticationError("Domain not found")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        name = self._full_name(name)

        try:
            self._get(
                "dns-add_record",
                query_params={"record": name, "type": rtype, "value": content},
            )
        except AlreadyExistError:
            pass

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        payload = self._get("dns-list_records")

        resource_list = payload.get("data", None)
        if not isinstance(resource_list, list):
            raise Exception(f"unable to get records: {payload}")

        resource_list = [
            resource for resource in resource_list if resource["zone"] == self.domain
        ]
        if rtype:
            resource_list = [
                resource for resource in resource_list if resource["type"] == rtype
            ]
        if name:
            name = self._full_name(name)
            resource_list = [
                resource for resource in resource_list if resource["record"] == name
            ]
        if content:
            resource_list = [
                resource for resource in resource_list if resource["value"] == content
            ]

        processed_records = []
        for dreamhost_record in resource_list:
            processed_records.append(
                {
                    "id": Provider._identifier(dreamhost_record),
                    "type": dreamhost_record["type"],
                    "name": dreamhost_record["record"],
                    "content": dreamhost_record["value"],
                }
            )

        return processed_records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            try:
                self.delete_record(identifier)
            except NonExistError:
                pass

        return self.create_record(rtype=rtype, name=name, content=content)

    # Delete existing records.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        to_deletes = []
        if identifier:
            record = Provider._id_to_record(identifier)
            to_deletes.append(record)
        else:
            records = self.list_records(rtype=rtype, name=name, content=content)
            to_deletes = records

        # for-loop to delete deletes.
        err = None
        for each in to_deletes:
            try:
                dreamhost_record = Provider._record_to_dreamhost_record(each)
                self._get("dns-remove_record", query_params=dreamhost_record)

            except Exception as exception:
                err = exception

            # Sleeping for a bit to avoid triggering ddos protecting in case of looped requests
            wait_time = int(self._get_provider_option("retry_wait_time") or 1000)
            time.sleep(wait_time / 1000)

        if err is not None:
            raise err

        return True

    # Helpers
    def _request(self, action="GET", url="", data=None, query_params=None):
        if data is None:
            data = {}

        if query_params is None:
            query_params = {}

        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        query_params["key"] = self._get_provider_option("auth_token")
        query_params["format"] = "json"
        if "cmd" not in query_params:
            query_params["cmd"] = url

        response = requests.request(
            action,
            self.api_endpoint,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        result = response.json()
        if result.get("result", "") != "success":
            err_msg = result.get("data", "")
            if err_msg in _DATA_NON_EXIST_ERROR_LIST:
                raise NonExistError(f"Dreamhost non-exist error: {result}")
            if err_msg in _DATA_ALREADY_EXIST_ERROR_LIST:
                raise AlreadyExistError(f"Dreamhost already-exist error: {result}")
            raise Exception(f"Dreamhost api error: {result}")
        return result



"""Module provider for GratisDNS"""

import logging
from argparse import ArgumentParser
from typing import List

import requests
from bs4 import BeautifulSoup

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for GratisDNS"""

    # XXX: Identifiers change on updates

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["gratisdns.dk"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify email address for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://admin.gratisdns.com"
        self.cookies = {}

    def authenticate(self):
        # Getting required cookie "ORGID"
        payload = {
            "action": "logmein",
            "login": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        # Cannot allow redirects, as we miss the cookie then
        response = requests.post(self.api_endpoint, data=payload, allow_redirects=False)
        response.raise_for_status()

        if "ORGID" not in response.cookies:
            raise AuthenticationError("Unexpected auth response")
        self.cookies["ORGID"] = response.cookies["ORGID"]

        # Make sure domain exists
        # domain is stored in self.domain from BaseProvider
        domains = self._list_domains()
        for domain in domains:
            if domain == self.domain:
                # Domain name is the ID
                self.domain_id = domain
                break
        else:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def cleanup(self) -> None:
        pass

    def _list_domains(self):
        query_params = {"action": "dns_primarydns"}
        response = self._get(query_params=query_params)
        html = BeautifulSoup(response.content, "html.parser")
        # NOTE: This could be more robust, by checking more of the tree
        domains = [x.contents[0] for x in html.find_all("th", scope="row")]
        LOGGER.debug("list_domains: %s", domains)

        return domains

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        query_params = {
            "action": "dns_primary_changeDNSsetup",
            "user_domain": self.domain_id,
        }
        response = self._get(query_params=query_params)
        html = BeautifulSoup(response.content, "html.parser")
        # List of records
        all_records = []
        # DNS records are grouped by type
        dns_record_types = html.find_all("div", "dns-records")
        for dns_record in dns_record_types:
            # Find the title of this group, aka record type
            record_type = dns_record.find("h2").contents[0].strip()
            # The entries themselves follow in the table below
            # Each row is a single entry
            entry_table_rows = dns_record.find("tbody").find_all("tr")
            for row in entry_table_rows:
                cols = row.find_all("td")
                # Format is; name, content, ttl, EDIT/DELETE button
                record_name = cols[0].contents[0]
                record_content = cols[1].contents[0]
                record_ttl = cols[2].contents[0]
                # We have to pull the ID from the button link
                record_id = None
                button_link = cols[3].find("a")
                if button_link:
                    start = button_link["href"].index("&id=") + 4
                    end = button_link["href"].index("&", start)
                    record_id = button_link["href"][start:end]

                processed_record = {
                    "type": record_type,
                    "name": record_name,
                    "ttl": record_ttl,
                    "content": record_content,
                    "id": record_id,
                }
                all_records.append(processed_record)

        records = self._filter_records(
            records=all_records, rtype=rtype, name=name, content=content
        )
        LOGGER.debug("list_records: %s", records)
        return records

    # Filter a list of records based on criteria
    # NOTE: Duplicate of lexicon/providers/transip.py _filter_records
    def _filter_records(self, records, rtype=None, name=None, content=None):
        _records = []
        for record in records:
            if (
                (not rtype or record["type"] == rtype)
                and (
                    not name or self._full_name(record["name"]) == self._full_name(name)
                )
                and (not content or record["content"] == content)
            ):
                _records.append(record)
        return _records

    def _get_content_entry(self, rtype):
        if rtype == "TXT":
            return "txtdata"
        if rtype == "A":
            return "ip"
        if rtype == "AAAA":
            return "ip"
        if rtype == "CNAME":
            return "cname"
        LOGGER.error("unknown record type")
        return None

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        # Figure out TTL
        ttl = 43200
        if self._get_lexicon_option("ttl"):
            ttl = self._get_lexicon_option("ttl")

        content_entry = self._get_content_entry(rtype)
        if content_entry is None:
            return False

        # Prepare query params and payload
        query_params = {
            "action": "dns_primary_record_add_" + rtype.lower(),
            "user_domain": self.domain_id,
        }
        payload = {
            "action": "dns_primary_record_added_" + rtype.lower(),
            "name": self._full_name(name),
            "ttl": ttl,
            content_entry: content,
            "user_domain": self.domain_id,
        }
        self._post(query_params=query_params, data=payload)
        # response = self._post(query_params=query_params, data=payload)
        # NOTE: We shouldn't return False on duplicate?
        #       The specification seems a little wage on this.
        # html = BeautifulSoup(response.content, "html.parser")
        # error = html.find('td', 'table-danger')
        # success = (error is None)

        LOGGER.debug("create_record: %s", True)
        return True

    def _lookup_record(self, identifier):
        # Pull all the records we know
        records = self.list_records()
        # Find the one with the provided identifier
        record = next((x for x in records if x["id"] == identifier), None)
        return record

    # Create or update a record.
    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        record = None
        # Try to find, either by identifier or by searching
        if identifier:
            # Find the relevant record by id
            record = self._lookup_record(identifier)
        else:
            # Find the relevant record by rtype and name
            records = self.list_records(rtype=rtype, name=name)
            if len(records) == 1:
                record = records[0]
                identifier = record["id"]
        # If none is found, create it
        if record is None:
            return self.create_record(rtype, name, content)

        # Update ttl if provided
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")
        # Update name if provided
        if name:
            record["name"] = self._full_name(name)
        if content:
            record["content"] = content

        content_entry = self._get_content_entry(rtype)
        if content_entry is None:
            return False

        # Prepare query params and payload
        query_params = {
            "action": "dns_primary_record_edit_" + rtype.lower(),
            "user_domain": self.domain_id,
            "id": identifier,
        }
        payload = {
            "action": "dns_primary_record_update_" + rtype.lower(),
            "id": identifier,
            "name": record["name"],
            "ttl": record["ttl"],
            content_entry: record["content"],
            "user_domain": self.domain_id,
        }
        response = self._post(query_params=query_params, data=payload)
        html = BeautifulSoup(response.content, "html.parser")
        error = html.find("td", "table-danger")
        success = error is None

        LOGGER.debug("update_record: %s", success)
        return success

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        success = True
        for record_id in delete_record_id:
            # Lookup the record to determine rtype
            if rtype is None:
                record = self._lookup_record(record_id)
                rtype = record["type"]

            # Prepare query params and payload
            query_params = {
                "action": "dns_primary_delete_" + rtype.lower(),
                "user_domain": self.domain_id,
                "id": record_id,
            }
            # _get is intentional, get really is used for deletion on gratisdns
            response = self._get(query_params=query_params)
            html = BeautifulSoup(response.content, "html.parser")
            error = html.find("td", "table-danger")
            success = success and (error is None)

        LOGGER.debug("delete_record: %s", success)
        return success

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            cookies=self.cookies,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response



"""Module provider for CloudXNS"""

import hashlib
import json
import logging
import time
from argparse import ArgumentParser
from typing import List
from urllib.parse import urlencode

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for CloudXNS"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["cloudxns.net"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify API-KEY for authentication"
        )
        parser.add_argument(
            "--auth-token", help="specify SECRET-KEY for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://www.cloudxns.net/api2"

    def authenticate(self):
        payload = self._get("/domain")
        for record in payload["data"]:
            if record["domain"] == self.domain + ".":
                self.domain_id = record["id"]
                break
        else:
            raise AuthenticationError("No domain found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        record = {
            "domain_id": self.domain_id,
            "host": self._relative_name(name),
            "value": content,
            "type": rtype,
            "line_id": 1,
        }
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        try:
            self._post("/record", record)
        except requests.exceptions.HTTPError as err:
            already_exists = err.response.json()["code"] == 34
            if not already_exists:
                raise

        # CloudXNS will return bad HTTP Status when error, will throw at
        # r.raise_for_status() in _request()
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        payload = self._get(
            "/record/" + self.domain_id, {"host_id": 0, "offset": 0, "row_num": 2000}
        )
        records = []
        for record in payload["data"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["host"]),
                "ttl": record["ttl"],
                "content": record["value"],
                # this id is useless unless your doing record linking. Lets return the
                # original record identifier.
                "id": record["record_id"],
            }
            if processed_record["type"] == "TXT":
                processed_record["content"] = processed_record["content"].replace(
                    '"', ""
                )
                # CloudXNS will add quotes automaticly for TXT records,
                # https://www.cloudxns.net/Support/detail/id/114.html
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            records = self.list_records(name=name)
            if len(records) == 1:
                identifier = records[0]["id"]
            else:
                raise Exception("Record identifier could not be found.")

        data = {
            "domain_id": self.domain_id,
            "host": self._relative_name(name),
            "value": content,
            "type": rtype,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        self._put("/record/" + identifier, data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete("/record/" + record_id + "/" + self.domain_id)

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        data["login_token"] = (
            self._get_provider_option("auth_username")
            + ","
            + self._get_provider_option("auth_token")
        )
        data["format"] = "json"
        if query_params:
            query_string = "?" + urlencode(query_params)
        else:
            query_string = ""
            query_params = {}
        if data:
            data = json.dumps(data)
        else:
            data = ""
        date = time.strftime("%a %b %d %H:%M:%S %Y", time.localtime())
        default_headers = {
            "API-KEY": self._get_provider_option("auth_username"),
            "API-REQUEST-DATE": date,
            "API-HMAC": hashlib.md5(
                f"{self._get_provider_option('auth_username')}{self.api_endpoint}{url}{query_string}{data}{date}{self._get_provider_option('auth_token')}".encode(
                    "utf-8"
                )
            ).hexdigest(),
            "API-FORMAT": "json",
        }
        default_auth = None
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Base provider module for all Lexicon providers"""

from __future__ import annotations

from abc import ABC, abstractmethod
from argparse import ArgumentParser
from re import Pattern
from typing import Any

from lexicon.config import ConfigResolver, legacy_config_resolver


class Provider(ABC):
    """
    This is the abstract class for all lexicon Providers.
    It provides common functionality and ensures that all implemented
    Providers follow a standard ducktype.
    All standardized options will be provided here as defaults, but can be overwritten
    by environmental variables and cli arguments.

    Common options are:

    action
    domain
    type
    name
    content
    ttl
    priority
    identifier

    The provider_env_cli_options will also contain any Provider specific options:

    auth_username
    auth_token
    auth_password
    ...

    :param config: is a ConfigResolver object that contains all the options
                   for this provider, merged from CLI and Env variables.
    """

    def __init__(self, config: ConfigResolver | dict[str, Any]):
        if not isinstance(config, ConfigResolver):
            # If config is a plain dict, we are in a legacy situation.
            # To protect the Provider API, the legacy dict is handled in a
            # correctly defined ConfigResolver.
            # Also, there may be some situation where `provider` key is not set in the config.
            # It should not happen when Lexicon is called from Client, as it will set itself
            # this key. However, there were no automated logic if the Provider is used directly.
            # So we provide this logic here.
            if not config.get("provider_name") and not config.get("provider"):
                config["provider_name"] = (
                    __name__  # Obviously we use the module name itself.
                )
            self.config = legacy_config_resolver(config)
        else:
            self.config = config

        # Default ttl
        self.config.with_dict({"ttl": 3600})

        self.provider_name = self.config.resolve(
            "lexicon:provider_name"
        ) or self.config.resolve("lexicon:provider")
        self.domain = str(self.config.resolve("lexicon:domain"))
        self.domain_id = None

    # Provider API: instance methods

    @abstractmethod
    def authenticate(self) -> None:
        """
        Authenticate against provider,
        Make any requests required to get the domain's id for this provider,
        so it can be used in subsequent calls.
        Should throw AuthenticationError or requests.HTTPError if authentication fails for any reason,
        of if the domain does not exist.
        """

    def cleanup(self) -> None:
        """
        Clean any relevant resource before this provider instance is closed.
        """

    @abstractmethod
    def create_record(self, rtype: str, name: str, content: str) -> bool:
        """
        Create record. If record already exists with the same content, do nothing.
        """

    @abstractmethod
    def list_records(
        self,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> list[dict[str, Any]]:
        """
        List all records. Return an empty list if no records found
        type, name and content are used to filter records.
        If possible filter during the query, otherwise filter after response is received.
        """

    @abstractmethod
    def update_record(
        self,
        identifier: str | None = None,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> bool:
        """
        Update a record. Identifier must be specified.
        """

    @abstractmethod
    def delete_record(
        self,
        identifier: str | None = None,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> bool:
        """
        Delete an existing record.
        If record does not exist, do nothing.
        If an identifier is specified, use it, otherwise do a lookup using type, name and content.
        """

    # Provider API: static methods

    @staticmethod
    @abstractmethod
    def get_nameservers() -> list[str] | list[Pattern]:
        """
        Return the list of nameservers for this DNS provider
        """

    @staticmethod
    @abstractmethod
    def configure_parser(parser: ArgumentParser) -> None:
        """
        Configure the given parser for the provider needs
        (e.g. specific CLI flags for auth)
        """

    # Helpers

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
    ) -> Any:
        """Execute an HTTP request against the DNS provider API"""
        raise NotImplementedError(
            "You must implement _request() to use _get()/_post()/_put()/_patch()/_delete() methods."
        )

    def _get(self, url: str = "/", query_params: dict[str, Any] | None = None) -> Any:
        return self._request("GET", url, query_params=query_params)

    def _post(
        self,
        url: str = "/",
        data: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
    ) -> Any:
        return self._request("POST", url, data=data, query_params=query_params)

    def _put(
        self,
        url: str = "/",
        data: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
    ) -> Any:
        return self._request("PUT", url, data=data, query_params=query_params)

    def _patch(
        self,
        url: str = "/",
        data: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
    ) -> Any:
        return self._request("PATCH", url, data=data, query_params=query_params)

    def _delete(
        self, url: str = "/", query_params: dict[str, Any] | None = None
    ) -> Any:
        return self._request("DELETE", url, query_params=query_params)

    def _fqdn_name(self, record_name: str) -> str:
        # strip trailing period from fqdn if present
        record_name = record_name.rstrip(".")
        # check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = f"{record_name}.{self.domain}"
        return f"{record_name}."  # return the fqdn name

    def _full_name(self, record_name: str) -> str:
        # strip trailing period from fqdn if present
        record_name = record_name.rstrip(".")
        # check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = f"{record_name}.{self.domain}"
        return record_name

    def _relative_name(self, record_name: str) -> str:
        # strip trailing period from fqdn if present
        record_name = record_name.rstrip(".")
        # check if the record_name is fully specified
        if record_name.endswith(self.domain):
            record_name = record_name[: -len(self.domain)]
            record_name = record_name.rstrip(".")
        return record_name

    def _clean_TXT_record(self, record: dict[str, Any]) -> dict[str, Any]:
        if record["type"] == "TXT":
            # Some providers have quotes around the TXT records,
            # so we're going to remove those extra quotes
            record["content"] = record["content"][1:-1]
        return record

    def _get_lexicon_option(self, option: str) -> str | None:
        return self.config.resolve(f"lexicon:{option}")

    def _get_provider_option(self, option: str) -> str | None:
        return self.config.resolve(f"lexicon:{self.provider_name}:{option}")



"""Module provider for Yandex PDD

API doc: https://yandex.com/dev/domain/doc/reference/dns-add.html
"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

__author__ = "Aliaksandr Kharkevich"
__license__ = "MIT"
__contact__ = "https://github.com/kharkevich"

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Yandex PDD"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["yandex.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token",
            help="specify PDD token (https://yandex.com/dev/domain/doc/concepts/access.html)",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://pddimp.yandex.ru/api2/admin/dns"

    def authenticate(self):
        payload = self._get(f"/list?domain={self.domain}")
        if payload["success"] != "ok":
            raise AuthenticationError("No domain found")
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        if rtype in ("CNAME", "MX", "NS"):
            # make sure a the data is always a FQDN for CNAMe.
            content = content.rstrip(".") + "."

        querystring = f"domain={self.domain_id}&type={rtype}&subdomain={self._relative_name(name)}&content={content}"
        if self._get_lexicon_option("ttl"):
            querystring += f"&ttl={self._get_lexicon_option('ttl')}"

        payload = self._post("/add", {}, querystring)

        return self._check_exitcode(payload, "create_record")

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        url = f"/list?domain={self.domain_id}"
        records = []
        payload = {}

        next_url = url
        while next_url is not None:
            payload = self._get(next_url)
            if (
                "links" in payload
                and "pages" in payload["links"]
                and "next" in payload["links"]["pages"]
            ):
                next_url = payload["links"]["pages"]["next"]
            else:
                next_url = None

            for record in payload["records"]:
                if record["type"] == "MX":
                    assembled_content = f"{record['priority']} {record['content']}"
                if record["type"] == "SRV":
                    if "target" in record:
                        srv_target = record["target"]
                    else:
                        srv_target = record["content"]
                    assembled_content = f"{record['priority']} {record['weight']} {record['port']} {srv_target}"
                else:
                    assembled_content = record.get("content")
                record_name = (
                    f"{record['subdomain']}.{self.domain_id}"
                    if record["subdomain"] != "@"
                    else self.domain_id
                )
                processed_record = {
                    "type": record["type"],
                    "name": record_name,
                    "ttl": record["ttl"],
                    "content": assembled_content,
                    "id": record["record_id"],
                }
                records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Just update existing record. If Identifier is not provided, update the latest entry with matching name and rtype.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            # get existing entries, and pick the last one for update
            records = self.list_records(rtype=rtype, name=name)
            if not records:
                return False
            identifier = records[-1]["id"]

        data = ""
        if rtype:
            data += f"&type={rtype}"
        if name:
            data += f"&subdomain={self._relative_name(name)}"
        if content:
            data += f"&content={content}"

        payload = self._post(
            "/edit", {}, f"domain={self.domain_id}&record_id={identifier}" + data
        )

        return self._check_exitcode(payload, "update_record")

    # Delete an existing record.
    # If record does not exist (I'll hope), do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._post("/del", {}, f"domain={self.domain_id}&record_id={record_id}")

        # return self._check_exitcode(payload, 'delete_record')
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "PddToken": self._get_provider_option("auth_token"),
        }

        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        return response.json()

    def _check_exitcode(self, payload, title):
        if payload["success"] == "ok":
            LOGGER.debug("%s: %s", title, payload["success"])
            return True
        if payload["error"] == "record_exists":
            LOGGER.debug("%s: %s", title, True)
            return True
        LOGGER.debug("%s: %s", title, payload["error"])
        return False



"""Module provider for PointHQ"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for PointHQ"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["pointhq.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify email address for authentication"
        )
        parser.add_argument("--auth-token", help="specify token for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://pointhq.com"

    def authenticate(self):
        payload = self._get(f"/zones/{self.domain}")

        if not payload["zone"]:
            raise AuthenticationError("No domain found")

        self.domain_id = payload["zone"]["id"]

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        # check if record already exists
        existing_records = self.list_records(rtype, name, content)
        if len(existing_records) == 1:
            return True

        payload = self._post(
            f"/zones/{self.domain_id}/records",
            {
                "zone_record": {
                    "record_type": rtype,
                    "name": self._relative_name(name),
                    "data": content,
                }
            },
        )

        LOGGER.debug("create_record: %s", payload["zone_record"])
        return bool(payload["zone_record"])

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        filter_query = {}
        if rtype:
            filter_query["record_type"] = rtype
        if name:
            filter_query["name"] = self._relative_name(name)

        payload = self._get(f"/zones/{self.domain_id}/records", filter_query)

        records = []
        for record in payload:
            processed_record = {
                "type": record["zone_record"]["record_type"],
                "name": self._full_name(record["zone_record"]["name"]),
                "ttl": record["zone_record"]["ttl"],
                "content": record["zone_record"]["data"],
                "id": record["zone_record"]["id"],
            }
            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        if rtype:
            data["record_type"] = rtype
        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["data"] = content

        payload = self._put(
            f"/zones/{self.domain_id}/records/{identifier}", {"zone_record": data}
        )

        LOGGER.debug("update_record: %s", payload)
        return bool(payload["zone_record"])

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/zones/{self.domain_id}/records/{record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            auth=requests.auth.HTTPBasicAuth(
                self._get_provider_option("auth_username"),
                self._get_provider_option("auth_token"),
            ),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for deSEC"""

import hashlib
import logging
import re
from argparse import ArgumentParser

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util.retry import Retry

from lexicon._private.discovery import lexicon_version
from lexicon.config import ConfigResolver
from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# Type aliases
RecordType = tuple[dict, int]
RecordList = dict[str, RecordType]
StrDict = dict[str, str]
StrDictList = list[StrDict]
OptStr = str | None
OptStrDict = StrDict | None
SanitizedResponseType = tuple[str, str, dict]


class Provider(BaseProvider):
    """Provider class for deSEC"""

    @staticmethod
    def get_nameservers() -> list[str]:
        return ["ns1.desec.io", "ns2.desec.org"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api token for authentication")
        parser.add_argument(
            "--auth-username", help="specify email address for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config: ConfigResolver):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://desec.io/api/v1"
        self._lexicon_version = lexicon_version()
        self._token = self._get_provider_option("auth_token")
        self._priority = self._get_lexicon_option("priority")
        if self._priority and not self._priority.isnumeric():
            raise ValueError(f"Priority argument '{self._priority}' is not numeric.")

        # RegEx patterns, priority optional
        self._re = {
            "MX": re.compile(r"((?P<priority>\d+)\s+)?(?P<target>.+)"),
            "SRV": re.compile(
                r"((?P<priority>\d+)\s+)?(?P<weight>\d+)\s+(?P<port>\d+)\s+(?P<target>.+)"
            ),
        }

        # deSEC enforces rate limits, which are hit by rapid successive requests,
        # like by the automated tests via pytest. The API responses with status 429
        # and a retry-after header, which we want to use for retries.

        # https://desec.readthedocs.io/en/latest/rate-limits.html
        # dns_api_per_domain_expensive: 2/s - 15/min - 100/h - 300/day
        self._session = requests.Session()
        self._session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(
                    allowed_methods=None,  # Allow all methods
                    respect_retry_after_header=True,
                )
            ),
        )

    def authenticate(self):
        # Handle authentication
        username = self._get_provider_option("auth_username")
        password = self._get_provider_option("auth_password")
        if self._token and (username or password):
            raise AuthenticationError("Multiple authentication mechanisms specified.")
        if username and password and not self._token:
            self._login(username, password)
        if not self._token:
            raise AuthenticationError("No valid authentication mechanism specified.")

        # Check if domain exists
        try:
            self._get("..")
            self.domain_id = self.domain
            LOGGER.debug(f"authenticate: domain '{self.domain}' found.")
        except HTTPError as err:
            raise AuthenticationError(
                f"Domain '{self.domain}' not found ({err.response.status_code})."
            )

    def cleanup(self) -> None:
        pass

    def list_records(
        self, rtype: OptStr = None, name: OptStr = None, content: OptStr = None
    ) -> list[dict]:
        rec_sets = self._fetch_record_sets(rtype, name, content)
        records = [
            {
                "type": rec_set["type"],
                "ttl": rec_set["ttl"],
                "name": self._format_name(rec_set),
                "id": identifier,
                "content": sanitized,
            }
            | options
            for identifier, (rec_set, index) in rec_sets.items()
            for sanitized, dirty, options in [
                self._sanitize_response_content(rec_set, index)
            ]
        ]
        LOGGER.debug("list_records: %s", records)
        return records

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype: str, name: str, content: str) -> bool:
        desec_rec, index = self._get_record_set(rtype, name)
        desec_content = self._sanitize_request_content(content, rtype)
        subname = self._relative_name(name) if name else ""
        if not desec_rec or index == -1:
            # Create new record set, as it doesn't exist yet
            LOGGER.debug("_create_record: Creating new record set")
            self._post(
                "",
                {
                    "records": [desec_content],
                    "type": rtype,
                    "subname": subname,
                },
            )
            return True

        # Check if record is in record set
        desec_records: list[str] = desec_rec["records"]
        if content in desec_records or desec_content in desec_records:
            LOGGER.debug("_create_record: The record already exists. Ignore.")
            return True

        # Patch existing record set
        desec_records.append(desec_content)
        self._patch(f"{subname}/{rtype}", desec_rec)
        return True

    # Update a record.
    def update_record(
        self,
        identifier: OptStr = None,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
    ) -> bool:
        if identifier and name:
            # We can only filter for record type, as the subname could differ
            rec_sets = self._fetch_record_sets(rtype)
            if identifier not in rec_sets:
                LOGGER.warning(
                    f"update_record: No match for identifier '{identifier}'. Abort."
                )
                return False
            desec_rec, index = rec_sets[identifier]
        else:
            # We can't filter for content, as it likely changed
            desec_rec, index = self._get_record_set(rtype, name, None, identifier)

        if not desec_rec or index == -1:
            LOGGER.warning("update_record: No matching record found. Abort.")
            return False

        desec_records: list[str] = desec_rec["records"]
        rtype = rtype or desec_rec["type"]
        content = content or desec_records[index]

        # The subname is immutable and can't be modified.
        # We need to delete the old record and create a new one.
        old_subname = desec_rec["subname"]
        new_subname = self._relative_name(name) if name else ""
        if old_subname != new_subname:
            LOGGER.debug(
                f"update_record: new subname '{new_subname}' differs from old '{old_subname}'. Delete and recreate record."
            )
            return self._delete_record(
                desec_rec, index, rtype, old_subname
            ) and self.create_record(rtype, new_subname, content)

        # Patch the content
        new_content = self._sanitize_request_content(content or "", rtype or "")
        desec_records[index] = new_content
        self._patch(f"{new_subname or '@'}/{rtype}", desec_rec)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(
        self,
        identifier: OptStr = None,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
    ) -> bool:
        # Get first item
        desec_rec, index = self._get_record_set(rtype, name, content, identifier)
        if not desec_rec or index == -1:
            LOGGER.debug("delete_record: Record not found. Ignore.")
            return True
        subname = desec_rec["subname"]
        rtype = rtype or desec_rec["type"]
        if rtype and name and not (content or identifier):
            LOGGER.debug(
                f"delete_record: remove whole '{rtype}' record set '{subname}'"
            )
            self._delete(f"{subname}/{rtype}/")
            return True
        return self._delete_record(desec_rec, index, rtype, subname)

    def _delete_record(
        self, desec_rec: dict, index: int, rtype: str, subname: str
    ) -> bool:
        # Delete specific record
        desec_records: list[str] = desec_rec["records"]
        removed = desec_records.pop(index)
        self._patch(f"{subname}/{rtype}", desec_rec)
        LOGGER.debug(
            f"_delete_record: removed '{rtype}' item from '{subname}' record set, with content '{removed}'"
        )
        return True

    # Helpers

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: OptStrDict = None,
        query_params: OptStrDict = None,
    ):
        # TTL is required for all deSEC record sets
        # It can be omitted in PATCH requests, which keeps the RR's current TTL
        if data:
            ttl_cur = int(data.get("ttl", 0))
            ttl_opt = int(self._get_lexicon_option("ttl") or 3600)
            if ttl_cur == ttl_opt or (action == "PATCH" and ttl_cur < 3600):
                # Preserve current value if under 3600, this is necessary for dynDNS use cases
                # The API clamps values under 3600 (if set) but keeps the current if unset
                del data["ttl"]
            else:
                # Override by user / set default
                data["ttl"] = str(ttl_opt)

        response = self._session.request(
            action,
            f"{self.api_endpoint}/domains/{self.domain}/rrsets/{url}/",
            params=query_params,
            json=data,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": f"lexicon/{self._lexicon_version} desec",
                "Authorization": f"Token {self._token}",
            },
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        # DELETE responses without a body
        if response.status_code == 204:
            return {}
        return response.json()

    def _login(self, username: str, password: str) -> None:
        LOGGER.debug("_login: logging in with username / password")
        auth_res = requests.post(
            self.api_endpoint + "/auth/login/",
            None,
            {"email": username, "password": password},
        )
        auth_res.raise_for_status()

        json_res = auth_res.json()
        if json_res.get("mfa"):
            raise AuthenticationError("Login with enabled MFA/2FA is not supported.")

        self._token = json_res.get("token")
        if not self._token:
            raise AuthenticationError("Login successful, but no token was acquired.")

    def _fetch_record_sets(
        self, rtype: OptStr = None, name: OptStr = None, content: OptStr = None
    ) -> RecordList:
        desec_content = self._sanitize_request_content(content or "", rtype or "")
        response = self._get(
            "",
            {
                "type": rtype or None,
                "subname": self._relative_name(name) if name else None,
            },
        )

        # Generates a dict with the identifier as key
        # and a tuple of the record set and the index of related record
        # {id: (rec_set, index)}
        id_sets = {
            self._identifier(rec_set, rec): (rec_set, i)
            for rec_set in response
            for i, rec in enumerate(rec_set["records"])
            if (not content or content == rec)
            or (not desec_content or desec_content == rec)
        }
        LOGGER.debug("_fetch_record_sets: %s", id_sets)
        return id_sets

    def _get_record_set(
        self,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
        identifier: OptStr = None,
    ) -> RecordType:
        rec_sets = self._fetch_record_sets(rtype, name, content)
        if not len(rec_sets):
            # No record set found
            LOGGER.debug("_get_record_set: No match.")
            return ({}, -1)

        rec_set, *_ = rec_sets.values()
        if identifier and identifier in rec_sets:
            # Return specified record set
            return rec_sets[identifier]

        # Return first result
        return rec_set

    # Override, handle apex
    def _relative_name(self, record_name: str) -> str:
        subname = super()._relative_name(record_name or "@")
        return subname if subname != "@" else ""

    # Override, allow foreign domains
    def _fqdn_name(self, record_name: str) -> str:
        return (
            record_name
            if record_name.endswith(".")
            else super()._fqdn_name(record_name)
        )

    @staticmethod
    def _format_name(match: StrDict) -> str:
        sub = match["subname"]
        return f"{'@.' if not sub else ''}{match['name']}".strip(".")

    @staticmethod
    def _identifier(match: StrDict, record: str) -> str:
        sha256 = hashlib.sha256()
        sha256.update(f"{match['name']} => {match['type']} => '{record}'".encode())
        return sha256.hexdigest()[0:7]

    def _sanitize_request_content(self, content: str, rtype: str) -> str:
        if not content:
            return ""
        match rtype:
            case "TXT":
                content = content.strip('"')
                return f'"{content}"' if content else ""
            case "CNAME":
                return self._fqdn_name(content)
            case "MX" | "SRV":
                # The priority is only relevant for MX and SRV types.
                # deSEC does not support this property, it is part of the record's content.
                parsed = self._parse_priority_record(content, rtype)
                priority = parsed.get("priority") or str(self._priority)
                parsed["priority"] = priority  # Ensure fallback for join operation
                if not priority:
                    raise ValueError("Priority value is not defined.")
                if self._priority and self._priority != priority:
                    raise ValueError(
                        f"The priority was specified as an argument ({self._priority}) "
                        f"and in the content ({priority}), but it doesn't match."
                    )
                return " ".join(parsed.values())
            case _:
                return content

    def _sanitize_response_content(
        self, rec_set: dict, index: int
    ) -> SanitizedResponseType:
        content = rec_set["records"][index]
        match rtype := rec_set["type"]:
            case "MX" | "SRV":
                parsed = self._parse_priority_record(content, rtype)
                if not (priority := parsed.get("priority")) or not priority.isnumeric():
                    raise Exception("Priority value is not present in content.")
                # Convert numeric options to int, see `technical_workbook.rst`
                options: dict = {
                    k: (int(v) if v.isnumeric() else v) for k, v in parsed.items()
                }
                return " ".join(parsed.values()), content, {rtype.lower(): options}
            case _:
                return content.strip('"'), content, {}

    def _parse_priority_record(self, content: str, rtype: str) -> StrDict:
        if not (match := self._re[rtype].match(content)):
            raise Exception(f"Content '{content}' is not valid for type '{rtype}'.")

        groups = match.groupdict()
        groups["target"] = self._fqdn_name(groups["target"])
        return groups
